#include <boost/asio.hpp> /// just including this requires to link with boost system :-(

#include <iostream>
#include <thread>
#include <string>
#include <functional>

/// LET's PUT THINGS TOGETHER
///         MessageProcessor  ==> each time a message comes in, post it as a task on the io service
///
///  PROBLEM : if no messages comes in, all tasks will finish and the io service "run()" will return
///            and as such no longer available when a new message does come in
/// SOLUTION : special construct to keep a service "busy"


namespace
{

class Message
{
public:
    explicit Message(const std::string& work) :
        mWork{work}
    {
    }

    void process()
    {
        std::cout << "processing : " << mWork << std::endl;
        std::this_thread::sleep_for(std::chrono::seconds(2));
        std::cout << "processed : " << mWork << std::endl;
        std::cout << std::endl;
    }

private:
    std::string mWork;
};


class MessageProcessor
{
public:
    MessageProcessor() :
        mIoService{},
        mThread{&MessageProcessor::threadBody, this}
    {
    }

    ~MessageProcessor()
    {
        stop();
        mThread.join();
    }

    void postMessage(const Message& message)
    {
        mIoService.post(std::bind(&Message::process, message));
    }

    void stop()
    {
        mIoService.stop();
    }

private:
    void threadBody()
    {
        boost::asio::io_service::work  work{mIoService}; /// the special construct to keep service busy
        mIoService.run();
    }

    boost::asio::io_service mIoService;
    std::thread mThread;
};

} // namespace

int main()
{
    MessageProcessor messageProcessor;


    Message m1{"work 1"};
    Message m2{"work 2"};
    Message m3{"work 3"};
    Message m4{"work 4"};
    Message m5{"work 5"};


    std::cout << "first 2 messages have arrived."  << std::endl;
    messageProcessor.postMessage(m1);
    messageProcessor.postMessage(m2);

    std::this_thread::sleep_for(std::chrono::seconds(1));
    std::cout << "1 second later another message have arrived."  << std::endl;
    messageProcessor.postMessage(m3);

    std::this_thread::sleep_for(std::chrono::seconds(10));
    std::cout << "10 seconds later 2 more messages have arrived."  << std::endl;
    messageProcessor.postMessage(m4);
    messageProcessor.postMessage(m5);

    std::this_thread::sleep_for(std::chrono::seconds(6));
    std::cout << "time to go home, stop please." << std::endl;
    messageProcessor.stop();

    std::cout << "all have finished, going home." << std::endl;
    std::cout <<  std::endl;
}
