#include <boost/asio.hpp> /// just including this requires to link with boost system :-(

#include <iostream>
#include <thread>
#include <string>
#include <functional>

/// what if an "EXCEPTION" occurs in the "TASK"
///  ==> escapes via the run ==> thread aborts
///
/// SOLUTION : catch the exception (and deal with it) and do the work/run thing again

namespace
{

class Message
{
public:
    explicit Message(const std::string& work) :
        mWork{work}
    {
    }

    void process()
    {
        std::cout << "processing : " << mWork << std::endl;
        if(mWork == "work 4")
        {
            throw "AUCH";
        }
        std::this_thread::sleep_for(std::chrono::seconds(2));
        std::cout << "processed : " << mWork << std::endl;
        std::cout << std::endl;
    }

private:
    std::string mWork;
};


class MessageProcessor
{
public:
    MessageProcessor() :
        mIoService{},
        mThread{&MessageProcessor::threadBody, this}
    {
    }

    ~MessageProcessor()
    {
        stop();
        mThread.join();
    }

    void postMessage(const Message& message)
    {
        mIoService.post(std::bind(&Message::process, message));
    }

    void stop()
    {
        mIoService.stop();
    }

private:
    void threadBody()
    {
        while(true)
        {
            try
            {
                boost::asio::io_service::work  work{mIoService}; /// the special construct to keep service busy
                mIoService.run();
                break;
            }
            catch(...)
            {
                std::cout << "CAUGHT something, let's recover." << std::endl;
            }
        }
    }

    boost::asio::io_service mIoService;
    std::thread mThread;
};

} // namespace

int main()
{
    MessageProcessor messageProcessor;


    Message m1{"work 1"};
    Message m2{"work 2"};
    Message m3{"work 3"};
    Message m4{"work 4"};
    Message m5{"work 5"};


    std::cout << "first 2 messages have arrived."  << std::endl;
    messageProcessor.postMessage(m1);
    messageProcessor.postMessage(m2);

    std::this_thread::sleep_for(std::chrono::seconds(1));
    std::cout << "1 second later another message have arrived."  << std::endl;
    messageProcessor.postMessage(m3);

    std::this_thread::sleep_for(std::chrono::seconds(10));
    std::cout << "10 seconds later another message have arrived."  << std::endl;
    messageProcessor.postMessage(m4);
    messageProcessor.postMessage(m5);

    std::this_thread::sleep_for(std::chrono::seconds(6));
    std::cout << "time to go home, stop please." << std::endl;
    messageProcessor.stop();

    std::cout << "all have finished, going home." << std::endl;
    std::cout <<  std::endl;
}
