#include <boost/asio.hpp>

#include <signal.h>
#include <boost/asio/signal_set.hpp>

#include <thread>
#include <iostream>
#include <string>

/// a server which is :
/// - ASYNCHRONOUS
/// - reads the message from the client
/// - sends this message back
/// - closes the connection

/// ===> scales, and if needed we can make it be served by multiple threads


namespace
{
std::atomic_bool stop{false};

std::size_t completionCondition(
        std::string& buffer,
        const boost::system::error_code& /*error*/, /// let's ignore
        std::size_t bytes_transferred)
{
    if(!bytes_transferred)
    {
        return 1;
    }
    return buffer[bytes_transferred - 1] == '\n' ? 0 : 1;
}

/// client session
struct Clientx
{
    explicit Clientx(boost::asio::io_context& ioContext) :
        mReadBuffer{},
        mSendBuffer{},
        mSocket{ioContext}
    {
    }
    ~Clientx()
    {
        std::cout << " a client session finished" << std::endl;
    }

    std::string mReadBuffer;
    std::string mSendBuffer;
    boost::asio::ip::tcp::socket mSocket;
};

using ClientSession = std::shared_ptr<Clientx>;

class Server
{
public:
    Server(
            boost::asio::io_context& ioContext,
            unsigned short service) :
        mIoContext{ioContext},
        mAcceptor{mIoContext, {boost::asio::ip::tcp::v4(), service}},
        mStop{false}
    {
    }

    void start()
    {
        auto newClient = std::make_shared<Clientx>(mIoContext);
        mAcceptor.async_accept(newClient->mSocket, std::bind(&Server::acceptHandler, this, newClient, std::placeholders::_1));
    }

    void stop()
    {
        mStop = true;
        mAcceptor.cancel();
    }

    void acceptHandler(
            ClientSession client,
            const boost::system::error_code& ec)
    {
        if(mStop)
        {
            return;
        }
        if(ec)
        {
            std::cout << "Accept failed : " << ec.message() << std::endl;
            std::cout << std::endl;
        }
        else
        {
            /// we will now launch the Asynchronous read (we use read  ==> to get entire message)

            boost::asio::async_read(
                    client->mSocket,
                    boost::asio::dynamic_buffer(client->mReadBuffer),
                    std::bind(completionCondition, std::ref(client->mReadBuffer), std::placeholders::_1, std::placeholders::_2),
                    std::bind(&Server::readHandler, this, client, std::placeholders::_1, std::placeholders::_2));
        }

        start();
    }

    void readHandler(
            ClientSession client,
            const boost::system::error_code& ec,
            size_t bytes)
    {
        if(mStop)
        {
            return;
        }
        if(ec)
        {
            std::cout << "Read failed : " << ec.message() << std::endl;
            std::cout << std::endl;
        }
        else
        {
            std::cout << "read : " << bytes << std::endl;
            client->mSendBuffer.resize(bytes);
            std::copy(client->mReadBuffer.cbegin(), client->mReadBuffer.cbegin() + bytes, client->mSendBuffer.begin());
            std::cout << client->mSendBuffer << std::endl;

            std::this_thread::sleep_for(std::chrono::seconds(5)); // processing takes time

            boost::asio::async_write(
                    client->mSocket,
                    boost::asio::buffer(client->mSendBuffer),
                    std::bind(&Server::writeHandler, this, client, std::placeholders::_1, std::placeholders::_2));
        }
    }

    void writeHandler(
            ClientSession /*client*/,
            const boost::system::error_code& ec,
            size_t /*bytes*/)
    {
        if(mStop)
        {
            return;
        }
        if(ec)
        {
            std::cout << "Write failed : " << ec.message() << std::endl;
            std::cout << std::endl;
        }
        /// we could choose to close the socket here, but since the session is going to end here
        /// the Client smart pointer is going to die and as such its socket member will get destructed ==> socket closed
    }

private:
    boost::asio::io_context& mIoContext;
    boost::asio::ip::tcp::acceptor mAcceptor;
    std::atomic_bool mStop;
};


void runServer(
        Server& server,
        boost::asio::io_context& ioContext)
{
    server.start();

    ioContext.run();
}

void handlerSignals(
        const boost::system::error_code& /*erc*/,
        int signal)
{
    std::cout << "signal " << signal << " occurred. Time to stop the application." <<std::endl;
    stop = true;
}

void asyncSignal(boost::asio::io_context& ioContext)
{
    boost::asio::signal_set signals{ioContext, SIGINT, SIGTERM};

    signals.async_wait(handlerSignals);
    ioContext.run();
}

} // namespace

int main()
{
    boost::asio::io_context ioContextSignal;
    boost::asio::io_context ioContextServer;

    std::thread t1{asyncSignal, std::ref(ioContextSignal)};

    Server server{ioContextServer, 8242};
    std::thread t2{runServer, std::ref(server), std::ref(ioContextServer)};

    t1.join(); /// wait on stop signal

    std::cout << "time stop go home, tell the server to stop" << std::endl;
    server.stop();

    t2.join();

    std::cout << "done" << std::endl;
}
