#include <boost/asio.hpp>

#include <signal.h>
#include <boost/asio/signal_set.hpp>

#define BOOST_COROUTINES_NO_DEPRECATION_WARNING
#include <boost/asio/spawn.hpp> /// this is now deprecated, but you could still use it ...
/// ==> this is the new library #include <boost/coroutine2/all.hpp> ===> but that is a different talk

#include <thread>
#include <iostream>
#include <string>

/// a server which is :
/// - ASYNCHRONOUS
/// - reads the message from the client
/// - sends this message back
/// - closes the connection

/// ===> scales, and if needed we can make it be served by multiple threads


namespace
{
std::atomic_bool stop{false};

std::size_t completionCondition(
        std::string& buffer,
        const boost::system::error_code& /*error*/, /// let's ignore
        std::size_t bytes_transferred)
{
    if(!bytes_transferred)
    {
        return 1;
    }
    return buffer[bytes_transferred - 1] == '\n' ? 0 : 1;
}

using ClientSocket = std::shared_ptr<boost::asio::ip::tcp::socket>;

class Server
{
public:
    Server(
            boost::asio::io_context& ioContext,
            unsigned short service) :
        mIoContext{ioContext},
        mAcceptor{mIoContext, {boost::asio::ip::tcp::v4(), service}},
        mStop{false}
    {
    }

    void start()
    {
        accept();
    }

    void stop()
    {
        mStop = true;
        mAcceptor.cancel();
    }

    void accept()
    {
        auto newClientSocket = std::make_shared<boost::asio::ip::tcp::socket>(mIoContext);
        mAcceptor.async_accept(*newClientSocket, std::bind(&Server::acceptHandler, this, newClientSocket, std::placeholders::_1));
    }

    void acceptHandler(
            ClientSocket socket,
            const boost::system::error_code& ec)
    {
        if(mStop)
        {
            return;
        }
        if(ec)
        {
            std::cout << "Accept failed : " << ec.message() << std::endl;
            std::cout << std::endl;
        }
        else
        {
            /// we will now launch the coroutine which deals with reading and responding
            boost::asio::spawn(
                    mIoContext,
                    std::bind(&Server::readWrite, this, socket, std::placeholders::_1));
        }

        accept();
    }

    void readWrite(
            ClientSocket socket,
            boost::asio::yield_context yield)
    {
        if(mStop)
        {
            return;
        }

        std::string readBuffer;
        boost::system::error_code ec{};
        const auto bytes = boost::asio::async_read(
                *socket,
                boost::asio::dynamic_buffer(readBuffer),
                std::bind(completionCondition, std::ref(readBuffer), std::placeholders::_1, std::placeholders::_2),
                yield[ec]);

        if(mStop)
        {
            return;
        }

        if(ec)
        {
            std::cout << "Read failed : " << ec.message() << std::endl;
            std::cout << std::endl;
            return;
        }

        std::string sendBuffer;
        std::cout << "read : " << bytes << std::endl;
        sendBuffer.resize(bytes);
        std::copy(readBuffer.cbegin(), readBuffer.cbegin() + bytes, sendBuffer.begin());
        std::cout << sendBuffer << std::endl;

        std::this_thread::sleep_for(std::chrono::seconds(5)); // processing takes time

        if(mStop)
        {
            return;
        }

        boost::asio::async_write(
                *socket,
                boost::asio::buffer(sendBuffer),
                yield[ec]);

        if(ec)
        {
            std::cout << "Write failed : " << ec.message() << std::endl;
            std::cout << std::endl;
        }
        std::cout << " a client session finished" << std::endl;
    }

private:
    boost::asio::io_context& mIoContext;
    boost::asio::ip::tcp::acceptor mAcceptor;
    std::atomic_bool mStop;
};

void handlerSignals(
        const boost::system::error_code& /*erc*/,
        int signal)
{
    std::cout << "signal " << signal << " occurred. Time to stop the application." <<std::endl;
    stop = true;
}

void asyncSignal(boost::asio::io_context& ioContext)
{
    boost::asio::signal_set signals{ioContext, SIGINT, SIGTERM};

    signals.async_wait(handlerSignals);
    ioContext.run();
}

} // namespace

int main()
{
    boost::asio::io_context ioContextSignal;
    boost::asio::io_context ioContextServer;

    std::thread t1{asyncSignal, std::ref(ioContextSignal)};

    Server server{ioContextServer, 8242};
    server.start();

    /// and now the team of threads that will run it

    std::vector<std::unique_ptr<std::thread>> mThreads;

    const int threads{3};

    for(auto thread = 0; thread < threads; ++thread)
    {
        mThreads.emplace_back(std::make_unique<std::thread>(
            [&ioContextServer]()
            {
                ioContextServer.run();
                std::cout << "worker thread finished"  << std::endl;
            } ));
    }

    t1.join(); /// wait on stop signal

    std::cout << "time stop go home, tell the server to stop" << std::endl;
    server.stop();   ///==> direct impact, no more connections are accepted

    ioContextServer.stop();  /// ==> have all pending async operations , aka tasks to be cancelled

    for(auto& thread : mThreads)
    {
        thread->join();
    }

    std::cout << "done" << std::endl;
}
