#include <boost/asio.hpp>

#include <signal.h>
#include <boost/asio/signal_set.hpp>

#include <thread>
#include <iostream>
#include <string>
#include <array>
#include <algorithm>

/// UDP : connectionLESS

/// a server which is :
/// - ASYNCHRONOUS
/// - reads the message from the client
/// - sends this message back

namespace
{
std::atomic_bool stop{false};

/// client session
struct Clientx
{
    explicit Clientx() :
        mReadBuffer{},
        mSendBuffer{},
        mSender{}
    {
    }
    ~Clientx()
    {
        std::cout << " a client session finished" << std::endl;
    }

    std::array<char, 70'000> mReadBuffer;
    std::array<char, 70'000> mSendBuffer;
    boost::asio::ip::udp::endpoint mSender;
};

using ClientSession = std::shared_ptr<Clientx>;

class Server
{
public:
    Server(
            boost::asio::io_context& ioContext,
            unsigned short service) :
        mIoContext{ioContext},
        mService{service},
        mSocket{mIoContext, {boost::asio::ip::udp::v4(), mService} },
        mStop{false}
    {
    }

    void start()
    {
        auto newClient = std::make_shared<Clientx>();
        mSocket.async_receive_from(
                boost::asio::buffer(newClient->mReadBuffer),
                newClient->mSender,
                std::bind(&Server::readHandler, this, newClient, std::placeholders::_1, std::placeholders::_2));
    }

    void readHandler(
            ClientSession client,
            const boost::system::error_code& ec,
            size_t bytes)
    {
        if(mStop)
        {
            return;
        }
        start(); /// here now, so we don't suffer from the 5 sec processing time

        if(ec)
        {
            std::cout << "Read failed : " << ec.message() << std::endl;
            std::cout << std::endl;
        }
        else
        {
            std::string message;
            message.resize(bytes);
            std::copy(client->mReadBuffer.cbegin(), client->mReadBuffer.cbegin() + bytes, message.begin());
            std::cout << "read : " << bytes << std::endl;
            std::cout << message << std::endl;

            std::copy(client->mReadBuffer.cbegin(), client->mReadBuffer.cbegin() + bytes, client->mSendBuffer.begin());

            std::this_thread::sleep_for(std::chrono::seconds(5)); // processing takes time

            mSocket.async_send_to(
                    boost::asio::buffer(client->mSendBuffer, bytes),
                    client->mSender,
                    std::bind(&Server::writeHandler, this, client, std::placeholders::_1, std::placeholders::_2));
        }
    }

    void writeHandler(
            ClientSession /*client*/,
            const boost::system::error_code& ec,
            size_t /*bytes*/)
    {
        if(mStop)
        {
            return;
        }
        if(ec)
        {
            std::cout << "Write failed : " << ec.message() << std::endl;
            std::cout << std::endl;
        }
    }

    void stop()
    {
        mStop = true;
        mSocket.cancel();
    }

private:
    boost::asio::io_context& mIoContext;
    unsigned short mService;
    boost::asio::ip::udp::socket mSocket;
    std::atomic_bool mStop;
};

void runServer(
        Server& server,
        boost::asio::io_context& ioContext)
{
    server.start();
    ioContext.run();
}

void handlerSignals(
        const boost::system::error_code& /*erc*/,
        int signal)
{
    std::cout << "signal " << signal << " occurred. Time to stop the application." <<std::endl;
    stop = true;
}

void asyncSignal(boost::asio::io_context& ioContext)
{
    boost::asio::signal_set signals{ioContext, SIGINT, SIGTERM};

    signals.async_wait(handlerSignals);
    ioContext.run();
}

} // namespace

int main()
{
    boost::asio::io_context ioContextSignal;
    boost::asio::io_context ioContextServer;

    std::thread t1{asyncSignal, std::ref(ioContextSignal)};

    Server server{ioContextServer, 9242};
    std::thread t2{runServer, std::ref(server), std::ref(ioContextServer)};

    t1.join(); /// wait on stop signal

    std::cout << "time stop go home, tell the server to stop" << std::endl;
    server.stop();

    t2.join();

    std::cout << "done" << std::endl;
}
