#include <boost/asio.hpp>

#include <signal.h>
#include <boost/asio/signal_set.hpp>

#include <thread>
#include <iostream>
#include <string>

namespace
{
std::atomic_bool stop{false};

class Server
{
public:
    Server(
            boost::asio::io_context& ioContext,
            unsigned short service) :
        mIoContext{ioContext},
        mAcceptor{mIoContext, {boost::asio::ip::tcp::v4(), service}},
        mStop{false}
    {
    }

    void serve() /// can only stop when a message has arrived and we loop again => drawback of synchronous
    {
        std::string message;
        while(!mStop)
        {
            message.clear();
            boost::asio::ip::tcp::socket sock{mIoContext};
            mAcceptor.accept(sock);
            if(mStop)
            {
                continue;
            }

            boost::system::error_code ec;
            boost::asio::read(
                    sock,
                    boost::asio::dynamic_buffer(message),
                    ec);
            if(ec == boost::asio::error::eof)
            {
                std::cout << "received message : " << message << std::endl;
            }

            if(mStop)
            {
                continue;
            }

            std::this_thread::sleep_for(std::chrono::seconds(5)); // processing takes time

            boost::asio::write(sock, boost::asio::buffer(message));
            /// shutdown writing on our socket so client knows entire message has arrived
            sock.shutdown(boost::asio::socket_base::shutdown_send);
        }
    }

    void stop()
    {
        mStop = true;
        mAcceptor.cancel();
    }

private:
    boost::asio::io_context& mIoContext;
    boost::asio::ip::tcp::acceptor mAcceptor;
    std::atomic_bool mStop;
};

void runServer(Server& server)
{
	server.serve();
}

void handlerSignals(
        const boost::system::error_code& /*erc*/,
        int signal)
{
    std::cout << "signal " << signal << " occurred. Time to stop the application." <<std::endl;
    stop = true;
}

void asyncSignal(boost::asio::io_context& ioContext)
{
    boost::asio::signal_set signals{ioContext, SIGINT, SIGTERM};

    signals.async_wait(handlerSignals);
    ioContext.run();
}

} // namespace

int main()
{
    boost::asio::io_context ioContextSignal;
    boost::asio::io_context ioContextServer;

    std::thread t1{asyncSignal, std::ref(ioContextSignal)};

    Server server{ioContextServer, 8242};
    std::thread t2{runServer, std::ref(server)};

    t1.join(); /// wait on stop signal

    std::cout << "time stop go home, tell the server to stop" << std::endl;
    server.stop();

    t2.join();

    std::cout << "done" << std::endl;
}
