#include <boost/asio.hpp>

#include <signal.h>
#include <boost/asio/signal_set.hpp>

#include <thread>
#include <iostream>
#include <sstream>
#include <array>

/// we read async and ignore what we read
/// we act as a black hole, /dev/null


namespace
{
std::atomic_bool stop{false};

struct Clientx
{
    explicit Clientx(boost::asio::io_context& ioContext) :
        mBuffer{},
        mSocket{ioContext}
    {
    }

    std::array<char, 1'000> mBuffer;
//    char mBuffer[1000];
    boost::asio::ip::tcp::socket mSocket;
};

using Client = std::shared_ptr<Clientx>;

class Acceptor
{
public:
    Acceptor(
            boost::asio::io_context& ioContext,
            unsigned short service) :
        mIoContext{ioContext},
        mAcceptor{mIoContext, {boost::asio::ip::tcp::v4(), service}},
        mStop{false}
    {
    }

    void start()
    {
        auto newClient = std::make_shared<Clientx>(mIoContext);
        mAcceptor.async_accept(newClient->mSocket, std::bind(&Acceptor::acceptHandler, this, newClient, std::placeholders::_1));
    }

    void stop()
    {
        mStop = true;
        mAcceptor.cancel();
    }
private:
    void acceptHandler(
            Client client,
            const boost::system::error_code& ec)
    {
        if(mStop)
        {
            return;
        }
        if(ec)
        {
            std::cout << "Accept failed : " << ec.message() << std::endl;
            std::cout << std::endl;
        }
        else
        {
            client->mSocket.async_read_some(boost::asio::buffer(client->mBuffer),
                        std::bind(&Acceptor::readHandler, this, client, std::placeholders::_1, std::placeholders::_2));
        }

        start();
    }

    void readHandler(
            Client client,
            const boost::system::error_code& ec,
            size_t bytes)
    {
        if(mStop)
        {
            return;
        }
        if(ec)
        {
            std::cout << "Read failed : " << ec.message() << std::endl;
            std::cout << std::endl;
        }
        else
        {
            std::cout << "read : " << bytes << std::endl;
            client->mSocket.async_read_some(boost::asio::buffer(client->mBuffer),
                        std::bind(&Acceptor::readHandler, this, client, std::placeholders::_1, std::placeholders::_2));
        }
    }

    boost::asio::io_context& mIoContext;
    boost::asio::ip::tcp::acceptor mAcceptor;
    std::atomic_bool mStop;
};


void handleConnections(
        Acceptor& acceptor,
        boost::asio::io_context& ioContext)
{
	acceptor.start();

	ioContext.run();
}

void handlerSignals(
        const boost::system::error_code& /*erc*/,
        int signal)
{
    std::cout << "signal " << signal << " occurred. Time to stop the application." <<std::endl;
    stop = true;
}

void asyncSignal(boost::asio::io_context& ioContext)
{
    boost::asio::signal_set signals{ioContext, SIGINT, SIGTERM};

    signals.async_wait(handlerSignals);
    ioContext.run();
}

} // namespace

int main()
{
    boost::asio::io_context ioContextSignal;
    boost::asio::io_context ioContextAcceptor;

    std::thread t1{asyncSignal, std::ref(ioContextSignal)};

    Acceptor acceptor{ioContextAcceptor, 8242};
    std::thread t2{handleConnections, std::ref(acceptor), std::ref(ioContextAcceptor)};

    t1.join(); /// wait on stop signal

    std::cout << "time stop go home, tell the acceptor to stop" << std::endl;
    acceptor.stop();

    t2.join();

    std::cout << "done" << std::endl;
}
