#include <boost/asio.hpp>

#include <signal.h>
#include <boost/asio/signal_set.hpp>

#include <thread>
#include <iostream>
#include <sstream>
#include <utility>

/// use Asio8-SendWrite1 (or 2) to send data to our little server here
/// we will read synchronously in this example
/// we need a buffer, but how big, we don't know how big the incoming message will be
///         ===> we use a buffer of 1'000'000 bytes, assuming this will be big enough ....

/// next problem : how do we know the message will be finished, what is the message delimiter (if any)
/// in this example we know that the aX-zX sequences are separated by a ";", let's consider that as a delimiter
/// (the real message delimiter was actually the '\n'), but we use ";" so we can easily think multiple messages have arrived
/// ====> read with a completion handler

/// note in case the would not find the delimiter within the 1'000'000 bytes, read will return with the buffer completely filled,
/// but not with what we consider a complete message

namespace
{
std::atomic_bool stop{false};

/// A return value of 0 indicates that the read operation is complete.
/// A non-zero return value indicates the maximum number of bytes to be read on the next call to the stream's read_some function.
std::size_t completionCondition(
        std::string& buffer,
        const boost::system::error_code& /*error*/, /// let's ignore
        std::size_t bytes_transferred)
{
    if(!bytes_transferred)
    {
        return 1;
    }
    return buffer[bytes_transferred - 1] == ';' ? 0 : 1;
}
/// ===> DRAWBACK : we progress 1 byte a time

class Acceptor
{
public:
    Acceptor(
            boost::asio::io_context& ioContext,
            unsigned short service) :
        mAcceptor{ioContext, {boost::asio::ip::tcp::v4(), service}},
        mStop{false}
    {
    }

    void start()
    {
        mAcceptor.async_accept(std::bind(&Acceptor::acceptHandler, this, std::placeholders::_1, std::placeholders::_2)); // the overload to pass in the endpoint is also possible
    }

    void stop()
    {
        mStop = true;
        mAcceptor.cancel();
    }
private:
    void acceptHandler(
            const boost::system::error_code& ec,
            boost::asio::ip::tcp::socket socket)
    {
        if(ec)
        {
            std::cout << "Accept failed : " << ec.message() << std::endl;
            std::cout << std::endl;
        }
        else
        {
            /// we will now synchronously read the data here (not such a brilliant idea, since we block other connections waiting for acceptance)

            {
                std::string message{};
                message.resize(1'000'000);
                const auto read = boost::asio::read(socket, boost::asio::buffer(message), std::bind(completionCondition, std::ref(message), std::placeholders::_1, std::placeholders::_2));

                std::cout << "read gave us : " << read << std::endl;
                std::cout << message << std::endl;
                std::cout << "       note the presence of our separator !!!" << std::endl;
            }

            std::cout << "let's read 2 bytes more: " << std::endl;  /// buffer now did not contain any extra bytes
            {
                std::string message{};
                message.resize(2);
                const auto read = boost::asio::read(socket, boost::asio::buffer(message));
                std::cout << "read gave us : " << read << std::endl;
                std::cout << message << std::endl;
            }

        }
        if(mStop)
        {
            return;
        }

        start();
    }

    boost::asio::ip::tcp::acceptor mAcceptor;
    std::atomic_bool mStop;
};


void handleConnections(
        Acceptor& acceptor,
        boost::asio::io_context& ioContext)
{
	acceptor.start();

	ioContext.run();
}

void handlerSignals(
        const boost::system::error_code& /*erc*/,
        int signal)
{
    std::cout << "signal " << signal << " occurred. Time to stop the application." <<std::endl;
    stop = true;
}

void asyncSignal(boost::asio::io_context& ioContext)
{
    boost::asio::signal_set signals{ioContext, SIGINT, SIGTERM};

    signals.async_wait(handlerSignals);
    ioContext.run();
}

} // namespace

int main()
{
    boost::asio::io_context ioContextSignal;
    boost::asio::io_context ioContextAcceptor;

    std::thread t1{asyncSignal, std::ref(ioContextSignal)};

    Acceptor acceptor{ioContextAcceptor, 8242};
    std::thread t2{handleConnections, std::ref(acceptor), std::ref(ioContextAcceptor)};

    t1.join(); /// wait on stop signal

    std::cout << "time stop go home, tell the acceptor to stop" << std::endl;
    acceptor.stop();

    t2.join();

    std::cout << "done" << std::endl;
}
