#include <boost/asio.hpp>

#include <signal.h>
#include <boost/asio/signal_set.hpp>

#include <thread>
#include <iostream>
#include <sstream>
#include <utility>
#include <array>
#include <string>

/// use Asio8-SendWrite1 (or 2) to send data to our little server here
/// we will read Asynchronously in this example
/// we need a buffer, but how big, we don't know how big the incoming message will be
///         ===> we use a big buffer and hope it is big enough  ===> not so nice, what if ...

/// next problem : how do we know the message will be finished, what is the message delimiter (if any)
/// in this example we know that the aX-zX sequences are separated by a ";", let's consider that as a delimiter
/// (the real message delimiter was actually the '\n'), but we use ";" so we can easily think multiple messages have arrived
/// ====> read with a completion handler

/// NEW CHALLENGE : we need to keep the following items alive, until we have received the entire message :
///     - socket
///     - buffer
///     - (counter)

/// the read handler takes the error code and the bytes read (not the socket, nor the buffer)

/// we will wrap socket and buffer in "Session" or "Client" class

namespace
{
std::atomic_bool stop{false};

/// A return value of 0 indicates that the read operation is complete.
/// A non-zero return value indicates the maximum number of bytes to be read on the next call to the stream's read_some function.
std::size_t completionCondition(
        std::array<char, 1'000>& buffer,
        const boost::system::error_code& /*error*/, /// let's ignore
        std::size_t bytes_transferred)
{
    if(!bytes_transferred)
    {
        return 1;
    }
    return buffer[bytes_transferred - 1] == ';' ? 0 : 1;
}
/// ===> DRAWBACK : we progress 1 byte a time

/// clien session
struct Clientx
{
    explicit Clientx(boost::asio::io_context& ioContext) :
        mBuffer{},
        mSocket{ioContext}
    {
    }

    std::array<char, 1'000> mBuffer;
    boost::asio::ip::tcp::socket mSocket;
};

using Client = std::shared_ptr<Clientx>;


class Acceptor
{
public:
    Acceptor(
            boost::asio::io_context& ioContext,
            unsigned short service) :
        mIoContext{ioContext},
        mAcceptor{mIoContext, {boost::asio::ip::tcp::v4(), service}},
        mStop{false}
    {
    }

    void start()
    {
        auto newClient = std::make_shared<Clientx>(mIoContext);
        mAcceptor.async_accept(newClient->mSocket, std::bind(&Acceptor::acceptHandler, this, newClient, std::placeholders::_1));
    }

    void stop()
    {
        mStop = true;
        mAcceptor.cancel();
    }
private:
    void acceptHandler(
            Client client,
            const boost::system::error_code& ec)
    {
        if(mStop)
        {
            return;
        }
        if(ec)
        {
            std::cout << "Accept failed : " << ec.message() << std::endl;
            std::cout << std::endl;
        }
        else
        {
            /// we will now launch the Asynchronous read (we use (async_)read  ==> to get entire message)

            /// !!! keep the objects alive

            boost::asio::async_read(
                    client->mSocket,
                    boost::asio::buffer(client->mBuffer),
                    std::bind(completionCondition, std::ref(client->mBuffer), std::placeholders::_1, std::placeholders::_2),
                    std::bind(&Acceptor::readHandler, this, client, std::placeholders::_1, std::placeholders::_2));
        }

        start();
    }

    void readHandler(
            Client client,
            const boost::system::error_code& ec,
            size_t bytes)
    {
        if(mStop)
        {
            return;
        }
        if(ec)
        {
            std::cout << "Read failed : " << ec.message() << std::endl;
            std::cout << std::endl;
        }
        else
        {
            std::cout << "read : " << bytes << std::endl;
            std::string message{};
            message.resize(bytes);
            std::copy(client->mBuffer.cbegin(), client->mBuffer.cbegin() + bytes, message.begin());
            std::cout << message << std::endl;
        }
    }

    boost::asio::io_context& mIoContext;
    boost::asio::ip::tcp::acceptor mAcceptor;
    std::atomic_bool mStop;
};

void handleConnections(
        Acceptor& acceptor,
        boost::asio::io_context& ioContext)
{
	acceptor.start();

	ioContext.run();
}

void handlerSignals(
        const boost::system::error_code& /*erc*/,
        int signal)
{
    std::cout << "signal " << signal << " occurred. Time to stop the application." <<std::endl;
    stop = true;
}

void asyncSignal(boost::asio::io_context& ioContext)
{
    boost::asio::signal_set signals{ioContext, SIGINT, SIGTERM};

    signals.async_wait(handlerSignals);
    ioContext.run();
}

} // namespace

int main()
{
    boost::asio::io_context ioContextSignal;
    boost::asio::io_context ioContextAcceptor;

    std::thread t1{asyncSignal, std::ref(ioContextSignal)};

    Acceptor acceptor{ioContextAcceptor, 8242};
    std::thread t2{handleConnections, std::ref(acceptor), std::ref(ioContextAcceptor)};

    t1.join(); /// wait on stop signal

    std::cout << "time stop go home, tell the acceptor to stop" << std::endl;
    acceptor.stop();

    t2.join();

    std::cout << "done" << std::endl;
}
